@@ -7,6 +7,7 @@ import pathlib
 import re
 import shutil
 from dataclasses import dataclass, field
+from functools import cache
 from typing import Dict
 from xml.dom.minidom import parse
 
@@ -16,6 +17,7 @@ DEFAULT_FIX_VERSION = "4.2"
 FIX_TAG_ID_SENDING_TIME = "52"
 FIX_TAG_ID_SENDER_COMP_ID = "49"
 FIX_TAG_ID_TARGET_COMP_ID = "56"
+SESSION_LEVEL_TAGS = ['8', '34', '9', '10']
 
 DEFAULT_CFG_FILE_PATH = os.environ["HOME"] + "/.fixations.ini"
 DEFAULT_DATA_DIR_PATH = os.environ["HOME"] + "/.fixations"
@@ -150,7 +152,7 @@ def extract_elements_from_file_by_tag_name(fix_version, file, tag_name):
     elements = doc.getElementsByTagName(tag_name)
     return elements
 
-
+@cache
 def extract_tag_dict_for_fix_version(fix_version=DEFAULT_FIX_VERSION):
     versions = get_list_of_available_fix_versions()
     assert fix_version in versions, f"The specified FIX version:{fix_version} is not valid. Use one of these {versions}"
@@ -159,21 +161,19 @@ def extract_tag_dict_for_fix_version(fix_version=DEFAULT_FIX_VERSION):
     fields = extract_elements_from_file_by_tag_name(fix_version, "Fields.xml", "Field")
     tag_dict_by_id = {}
     for field in fields:
-        id, name, type, desc = extract_tag_data_from_xml_field(field)
-        #        tag_dict_by_id[id] = {'id': id, 'name': name, 'type': type, 'desc': desc, 'values': {}}
-        tag_dict_by_id[id] = FixTag(id, name, type, desc, {})
+        tag_id, name, tag_type, desc = extract_tag_data_from_xml_field(field)
+        tag_dict_by_id[tag_id] = FixTag(tag_id, name, tag_type, desc, {})
 
     # Extract all FIX tag values from Enums XML file and attach them to the tag dictionary
     enums = extract_elements_from_file_by_tag_name(fix_version, "Enums.xml", "Enum")
     for enum in enums:
-        id, name, value, desc = extract_tag_data_from_xml_enum(enum)
-        #        tag_dict_by_id[id]['values'][value] = {'value': value, 'name': name, 'desc': desc}
+        tag_id, name, value, desc = extract_tag_data_from_xml_enum(enum)
         fix_tag_value = FixTagValue(value, name, desc)
-        if id in tag_dict_by_id:
-            tag_dict_by_id[id].values[value] = fix_tag_value
+        if tag_id in tag_dict_by_id:
+            tag_dict_by_id[tag_id].values[value] = fix_tag_value
         else:
             # Somehow the quality of the data is not that great and some enum values reference tags that don't exist
-#            print(f"ERROR: id:{id} for name:{name}, value:{value}, desc:{desc} doesn't exist")
+            #            print(f"ERROR: id:{id} for name:{name}, value:{value}, desc:{desc} doesn't exist")
             pass
 
     return tag_dict_by_id
@@ -196,9 +196,117 @@ def save_tag_dict_to_json_file(tag_dict, json_file):
     with open(json_file, 'w') as json_file_fd:
         json_file_fd.write(json_object)
 
-    # -- Configuration --------
+
+def determine_fix_version(str_fix_lines):
+    for line in str_fix_lines:
+        match = re.search(r"8=FIX\.([.0-9]+)", line)
+        if match:
+            version = match.group(1)
+            return version
+
+    return None
 
 
+def get_fix_tag_dict_for_lines(str_fix_lines):
+    version = determine_fix_version(str_fix_lines)
+    assert version, "ERROR: can't extract FIX version from lines starting with line:{str_fix_lines[0]}"
+    fix_tag_dict = extract_tag_dict_for_fix_version(version)
+
+    return fix_tag_dict
+
+
+def parse_fix_line_into_kvs(line, fix_tag_dict):
+    match = re.search(r"8=FIX\.([.0-9]+)", line)
+    if not match:
+        return None
+
+    fix_start, fix_end = match.span()
+    body_length_start = line.find('9=')
+    separator = line[fix_end:body_length_start]
+
+    fix_line = line[fix_start:]
+    kv_parts = fix_line.split(separator)
+    kvs = {}
+    for kv_part in kv_parts:
+        if kv_part:
+            kv = re.search("^(\d+)=(.*)", kv_part)
+            if kv:
+                tag_id, value = kv.group(1, 2)
+                if tag_id in fix_tag_dict and value in fix_tag_dict[tag_id].values:
+                    value = f"{value} ({fix_tag_dict[tag_id].values[value].name})"
+                kvs[tag_id] = value
+            else:
+                print(f"ERROR: can't tokenize:'{kv_part}' into a key=value pair using separator:'{separator}'")
+    #    print(f"{fix_line}:\n\t{kvs}")
+    return kvs
+
+
+def extract_fix_lines_from_str_lines(str_fix_lines):
+    if len(str_fix_lines) and len(str_fix_lines[0].strip()) > 0:
+        version = determine_fix_version(str_fix_lines)
+        fix_tag_dict = extract_tag_dict_for_fix_version(version)
+
+        used_fix_tags = {}
+        fix_lines = []
+        for line in str_fix_lines:
+            fix_tags = parse_fix_line_into_kvs(line.strip(), fix_tag_dict)
+            if fix_tags:
+                if FIX_TAG_ID_SENDING_TIME in fix_tags:
+                    for fix_tag_key in fix_tags.keys():
+                        used_fix_tags[fix_tag_key] = 1
+                    fix_lines.append(fix_tags)
+                else:
+                    print(f"ERROR: FIX line w/o SENDING_TIME tag(52): {line}")
+
+        return fix_tag_dict, fix_lines, used_fix_tags
+    else:
+        return {}, [], {}
+
+
+def create_header_for_fix_lines(fix_lines, show_date):
+    headers = ['TAG_ID', 'TAG_NAME']
+    for fix_tags in fix_lines:
+        datetime = fix_tags[FIX_TAG_ID_SENDING_TIME]
+        if show_date is False:
+            datetime = remove_date_from_datetime(datetime)
+        headers.append(datetime)
+
+    return headers
+
+
+def create_fix_lines_grid(fix_tag_dict, fix_lines, used_fix_tags,
+                          with_session_level_tags=True, top_header_tags=[], show_date=False):
+    rows = []
+    for fix_tag in (*top_header_tags, *sorted(used_fix_tags, key=lambda k: int(k))):
+        if fix_tag in SESSION_LEVEL_TAGS and with_session_level_tags is False:
+            continue
+        if fix_tag in fix_tag_dict:
+            fix_tag_name = fix_tag_dict[fix_tag].name
+        else:
+            fix_tag_name = '???'
+        cols = [fix_tag, fix_tag_name]
+        for fix_tags in fix_lines:
+            value = fix_tags[fix_tag] if fix_tag in fix_tags else ''
+            if 'time' in fix_tag_name.lower() and show_date is False:
+                value = remove_date_from_datetime(value)
+            cols.append(value)
+        rows.append(cols)
+
+    headers = create_header_for_fix_lines(fix_lines, show_date)
+
+    return headers, rows
+
+
+def remove_date_from_datetime(dt_str):
+    # assume that the format is ISO 8601-ish and strip anything before the hh:mm:ss
+    found_timestamp = re.search(r'(\d+:\d+:\d+.*)', dt_str)
+    if found_timestamp:
+        return found_timestamp.group(1)
+    else:
+        return dt_str
+
+
+# -- Configuration --------
 cfg = configparser.ConfigParser()
 cfg_init()
 
